/*
 *  aatv - a program to watch TV on a text-based console
 *  Copyright (C) 2000 Florent de Lamotte
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *  
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *  Picked some code and information from 
 *  XawTV (c) 1998-2000 Gerd Knorr 
 *  you can download it at http://www.strusel007.de/linux/xawtv
 */

#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <aalib.h>

#include <asm/types.h>
#include <linux/videodev.h>

#include "aatv.h"

/* grabber device variables */
static struct video_capability	grab_cap;
static struct video_mmap	grab_buf [2];
static struct video_window	grab_win;
static struct video_audio 	grab_audio;
static int			grab_fd, grab_size, have_mmap;
static int			mix;
static int			current_buffer;
static unsigned char            *grab_data [2];
static struct video_channel	grab_chan;

/* still used in aatv.c */
struct video_picture	grab_pict;

void
grab_init (void)
{
     	int devmask;
	struct video_mbuf buf_info; 
	
     	if ((grab_fd = open(grab_device,O_RDWR)) == -1) {
	  	fprintf(stderr,"open %s: %s\n",grab_device,strerror(errno));
	  	exit(1);
     	}
     	if (ioctl(grab_fd,VIDIOCGCAP,&grab_cap) == -1) {
	  	fprintf(stderr,"%s: no v4l device\n",grab_device);
	  	exit(1);
     	}

     	/* set image source and TV norm */
     	grab_chan.channel = grab_input;
     	if (ioctl(grab_fd,VIDIOCGCHAN,&grab_chan)==-1) {
	  	perror("ioctl VIDIOCGCHAN");
	  	exit(1);
     	}
     	grab_chan.channel = grab_input;
     	grab_chan.norm    = grab_norm;
     	if (ioctl(grab_fd,VIDIOCSCHAN,&grab_chan)==-1) {
	  	perror("ioctl VIDIOCSCHAN");
	  	exit(1);
     	}

     	/* open the mixer */
     	if ((mix = open("/dev/mixer", O_RDONLY)) == -1) {
	  	perror("mixer open");
     	}
     	fcntl(mix,F_SETFD,FD_CLOEXEC);

     	if (-1 == ioctl(mix,MIXER_READ(SOUND_MIXER_DEVMASK),&devmask)) {
	  	perror("mixer read devmask");
     	}
     
     	close(mix);

     	/* turn audio on */
     	if (ioctl(grab_fd,VIDIOCGAUDIO,&grab_audio) == -1) {
	  	perror("ioctl VIDIOCGAUDIO");
	  	exit(1);
     	}
    
    
     	grab_audio.flags &= ~VIDEO_AUDIO_MUTE;

     	if (ioctl(grab_fd,VIDIOCSAUDIO,&grab_audio) == -1) {
	  	perror("ioctl VIDIOCSAUDIO");
	  	exit(1);
     	}


     	/* try to setup mmap-based capture */

	/* query buffer infos */
	if (ioctl(grab_fd, VIDIOCGMBUF, &buf_info) == -1) {
		perror ("ioctl VIDIOCGMBUF");
		exit (1);
	}

	current_buffer = 0;	

	/* First buffer */
     	grab_buf[0].format = VIDEO_PALETTE_GREY;
     	grab_buf[0].frame  = 0;
     	grab_buf[0].width  = grab_width;
     	grab_buf[0].height = grab_height;

	/* Second buffer */
     	grab_buf[1].format = VIDEO_PALETTE_GREY;
     	grab_buf[1].frame  = 1;
     	grab_buf[1].width  = grab_width;
     	grab_buf[1].height = grab_height;

	grab_data[0] = mmap(0,(size_t)buf_info.size ,PROT_READ|PROT_WRITE,MAP_SHARED,grab_fd,0);
	grab_data[1] = grab_data[0] + buf_info.offsets[1];
	
	if ((int)grab_data[0] != -1) { /* mmap succeded */
	  	have_mmap = 1;
		ioctl(grab_fd,VIDIOCMCAPTURE,&grab_buf[1]);
		return;
     	}

     	/* fallback to read() */
     	fprintf(stderr,"no mmap support available, using read()\n");
     	have_mmap = 0;
     	grab_pict.depth   = 8;
     	grab_pict.palette = VIDEO_PALETTE_GREY;
     	if (ioctl(grab_fd,VIDIOCSPICT,&grab_pict) == -1) {
	  	perror("ioctl VIDIOCSPICT");
	  	exit(1);
     	}
     	if (ioctl(grab_fd,VIDIOCGPICT,&grab_pict) == -1) {
	  	perror("ioctl VIDIOCGPICT");
	  	exit(1);
     	}
     	memset(&grab_win,0,sizeof(struct video_window));
     	grab_win.width  = grab_width;
     	grab_win.height = grab_height;
     	if (ioctl(grab_fd,VIDIOCSWIN,&grab_win) == -1) {
	  	perror("ioctl VIDIOCSWIN");
	  	exit(1);
     	}
     	if (ioctl(grab_fd,VIDIOCGWIN,&grab_win) == -1) {
	  	perror("ioctl VIDIOCGWIN");
	  	exit(1);
     	}

     	grab_size = grab_win.width * grab_win.height;
     	grab_data [current_buffer] = malloc((size_t)grab_size);
}

void
grab_free (void)
{
     	/* turn off audio */
     	grab_audio.flags |= VIDEO_AUDIO_MUTE;
     	if (ioctl(grab_fd,VIDIOCSAUDIO,&grab_audio) == -1) 
	  	perror("ioctl VIDIOCSAUDIO");
}

unsigned char*
grab_one(int *width, int *height)
{
     	int rc;
     	for (;;) {
		if (have_mmap) {
			if (ioctl(grab_fd,VIDIOCMCAPTURE,&grab_buf[current_buffer]) == -1) {
		    		perror("ioctl VIDIOCMCAPTURE");
	       		} else {
				current_buffer = (current_buffer + 1)%2;
		    		if (ioctl(grab_fd,VIDIOCSYNC,&grab_buf[current_buffer]) == -1) {
			 		perror("ioctl VIDIOCSYNC");
		    		} else {
			 		*width  = grab_buf[current_buffer].width;
			 		*height = grab_buf[current_buffer].height;
			 		return grab_data[current_buffer];
		    		}
	       		}
	  	} else {
	       		rc = read(grab_fd,grab_data[current_buffer],(size_t)grab_size);
	       		if (grab_size != rc) {
		    		fprintf(stderr,"grabber read error (rc=%d)\n",rc);
		    		return NULL;
	       		} else {
		    		*width  = grab_win.width;
		    		*height = grab_win.height;
		    		return grab_data[current_buffer];
	       		}
	  	}
     	}
}

unsigned long
get_freq (void)
{
	unsigned long freq;
     	if (ioctl(grab_fd, VIDIOCGFREQ, &freq) == -1)
	  	perror("ioctl VIDIOCGFREQ");
     	return (freq * 125 / 2);
}
     
int
change_freq (unsigned long freq)
{
     	freq = freq * 2 / 125;
     	if (ioctl(grab_fd, VIDIOCSFREQ, &freq) == -1)
	  	perror("ioctl VIDIOCSFREQ");
     	return 0;
}

int
set_video_norm (int vidnorm)
{
	grab_chan.norm = vidnorm;
	if (ioctl(grab_fd,VIDIOCSCHAN, &grab_chan)) {
		perror ("ioctl VIDIOCSCHAN");
		return -1;	
	}
	return 0;
}

int
get_video_norm (void)
{
	return grab_chan.norm;
}

int
set_video_source (int source)
{
	grab_chan.channel = source;
	if (ioctl(grab_fd,VIDIOCSCHAN,&grab_chan) == -1) {
		perror ("ioctl VIDIOCSCHAN");
		return -1;	
	}
	return 0;
}

int
get_video_source (void)
{
	return grab_chan.channel;
}

void
set_picture_properties (void)
{
     	if (ioctl(grab_fd, VIDIOCSPICT, &grab_pict) == -1)
	  	perror ("ioctl VIDIOSPICT");
     	if (ioctl(grab_fd, VIDIOCGPICT, &grab_pict) == -1)
	  	perror ("ioctl VIDIOSPICT");
}

void
get_picture_properties (void)
{
     	if (ioctl(grab_fd, VIDIOCGPICT, &grab_pict) == -1)
	  	perror ("ioctl VIDIOSPICT");
}
     
